package net.darkmist.defcon18;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.ref.WeakReference;
import java.nio.charset.Charset;
import java.security.SecureRandom;
import java.util.Random;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class Util
{
	private static final Class<Util> CLASS = Util.class;
	@SuppressWarnings("unused")
	private static final Logger logger = LoggerFactory.getLogger(CLASS);
	private static final Charset UTF8 = Charset.forName("UTF-8");
	private static final int NUM_RAND_TOKEN_LONGS = 4;
	private static final char[] nibble2hex = new char[]
	{
		'0', '1', '2', '3', '4', '5', '6', '7',
		'8', '9', 'a', 'b', 'c', 'd', 'e', 'f'
	};
	private static final byte nibble_mask = (byte)(0xf);	// no promotion to int...
	private static final byte nibble_width = (byte)(0x4);	// no promotion to int...
	public static final byte[] EMPTY_BYTE_ARRAY = new byte[0];
	public static final String[] EMPTY_STRING_ARRAY = new String[0];
	
	public static final String toHexString(byte[] bytes)
	{
		StringBuilder sb = new StringBuilder(bytes.length*2);

		for(byte b:bytes)
		{
			sb.append(nibble2hex[(b>>nibble_width)&nibble_mask]);
			sb.append(nibble2hex[b&nibble_mask]);
		}
		return sb.toString();
	}

	@SuppressWarnings("unchecked")
	public static String getStringProp(Class cls, String name, String def)
	{
		String val;

		if(name == null)
			throw new IllegalArgumentException("null property name passed");
		name = cls.getName() + '.' + name;
		if((val = System.getProperty(name))==null)
			return def;
		return val;
	}

	@SuppressWarnings("unchecked")
	public static long getLongProp(Class cls, String name, long def)
	{
		String val;

		if(name == null)
			throw new IllegalArgumentException("null property name passed");
		name = cls.getName() + '.' + name;
		if((val = System.getProperty(name))==null)
			return def;
		try
		{
			return Long.parseLong(val);
		}
		catch(NumberFormatException e)
		{
			throw new IllegalArgumentException("proeprty " + name + " not parsable as long (was =" + val + "=)", e);
		}
	}

	public static int getIntProp(Class<Object> cls, String name, int def)
	{
		String val;

		if(name == null)
			throw new IllegalArgumentException("null property name passed");
		name = cls.getName() + '.' + name;
		if((val = System.getProperty(name))==null)
			return def;
		try
		{
			return Integer.parseInt(val);
		}
		catch(NumberFormatException e)
		{
			throw new IllegalArgumentException("proeprty " + name + " not parsable as Integer (was =" + val + "=)", e);
		}
	}

	public static void sleep(long sleep)
	{
		long end;

		end = sleep + System.currentTimeMillis();
		do
		{
			if((sleep = end - System.currentTimeMillis())<=0)
				return;
			try
			{
				Thread.sleep(sleep);
			}
			catch(InterruptedException e)
			{
				throw new IllegalStateException("Thread unexpectedly interrupted during sleep.", e);
			}
		} while(true);
	}

	public static Random getRandom()
	{
		return new SecureRandom();
	}

	public static String randToken()
	{
		StringBuilder sb = new StringBuilder();
		Random rand = getRandom();

		for(int i=NUM_RAND_TOKEN_LONGS;i>0;i--)
			sb.append(Long.toHexString(rand.nextLong()));
		return sb.toString();
	}

	public static byte[] gzip(byte[] bytes)
	{
		try
		{
			ByteArrayOutputStream baos = new ByteArrayOutputStream(bytes.length);
			GZIPOutputStream gzip = new GZIPOutputStream(baos);
	
			gzip.write(bytes);
			gzip.close();
			return baos.toByteArray();
		}
		catch(IOException e)
		{
			throw new IllegalStateException("IOException writing to a ByteArrayOutputStream", e);
		}
	}

	public static byte[] zcat(byte[] bytes)
	{
		try
		{
			ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
			GZIPInputStream gzip = new GZIPInputStream(bais);
			return IOUtils.toByteArray(gzip);
		}
		catch(IOException e)
		{
			throw new IllegalStateException("IOException reading from a ByteArrayInputStream", e);
		}
	}

	public static String base64(byte[] bytes)
	{
		return toUTF8String(Base64.encodeBase64(bytes));
	}

	public static String base64(String str)
	{
		return base64(toUTF8Bytes(str));
	}

	public static byte[] unbase64(String str)
	{
		return Base64.decodeBase64(str);
	}

	public static byte[] decode(String str)
	{
		return zcat(unbase64(str));
	}

	public static String encode(byte[] bytes)
	{
		return base64(gzip(bytes));
	}

	public static String encode(String str)
	{
		return encode(toUTF8Bytes(str));
	}

	public static String decodeToString(String str)
	{
		return toUTF8String(decode(str));
	}

	public static String toUTF8String(byte[] bytes)
	{
		return new String(bytes, UTF8);
	}

	public static byte[] toUTF8Bytes(String str)
	{
		return str.getBytes(UTF8);
	}

	@SuppressWarnings("unchecked")
	public static String getResourceString(Class cls, String ext)
	{
		return new String(getResource(cls,ext));
	}

	@SuppressWarnings("unchecked")
	public static CharSequence getResourceStringRef(final Class cls, final String ext)
	{
		return new CharSequence()
		{
			private WeakReference<String> ref = null;

			public char charAt(int index)
			{
				return toString().charAt(index);
			}

			public int length()
			{
				return toString().length();
			}

			public CharSequence subSequence(int start, int end)
			{
				return toString().subSequence(start, end);
			}

			public synchronized String toString()
			{
				String str;

				if(ref == null || (str = ref.get())==null)
				{
					str = getResourceString(cls,ext);
					ref = new WeakReference<String>(str);
				}
				return str;
			}
		};
	}

	@SuppressWarnings("unchecked")
	public static byte[] getResource(Class cls, String ext)
	{
		InputStream in;
		String name;

		name = cls.getSimpleName() + '.' + ext;
		if((in = cls.getResourceAsStream(name))==null)
			throw new IllegalStateException("Unable to open resouce " + name);
		try
		{
			return IOUtils.toByteArray(in);
		}
		catch(IOException e)
		{
			throw new IllegalStateException("Unable to read resource " + name, e);
		}
		finally
		{
			IOUtils.closeQuietly(in);
		}
	}

	public static String ifNull(Object o, String ifNull)
	{
		return o == null ? ifNull : o.toString();
	}

	public static String ifNull(Object o)
	{
		return ifNull(o, "(null)");
	}

	public static StringBuilder join(StringBuilder sb, String seperator, Object...args)
	{
		if(args == null)
			return sb;
		switch(args.length)
		{
			case 0:
				return sb;
			case 1:
				return sb.append(ifNull(args[0]));
		}
		sb.append(ifNull(args[0]));
		for(int c=1;c<args.length;c++)
			sb.append(seperator).append(ifNull(args[c]));
		return sb;
	}

	public static String join(String seperator, Object...args)
	{
		return join(new StringBuilder(), seperator, args).toString();
	}
}
